# Amethyst - An experimental Minecraft server written in Elixir.
# Copyright (C) 2024 KodiCraft
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

defmodule Amethyst.ConnectionState.Play do
  require Amethyst.ConnectionState.Macros
  alias Amethyst.ConnectionState.Macros

  require Logger

  @moduledoc """
  This module contains the packets and logic for the Play state.
  """

  Macros.defpacket_clientbound :disconnect, 0x1D, 767, [reason: :nbt]
  Macros.defpacket_clientbound :login, 0x2B, 767, [
    entity_id: :int,
    is_hardcore: :bool,
    dimensions: {:array, [name: :string]},
    max_players: :varint,
    view_distance: :varint,
    simulation_distance: :varint,
    reduced_debug_info: :bool,
    enable_respawn_screen: :bool,
    do_limited_crafting: :bool,
    dimension_type: :varint,
    dimension_name: :string,
    hashed_seed: :long,
    game_mode: :ubyte,
    previous_game_mode: :byte,
    is_debug: :bool,
    is_flat: :bool,
    death_location: {:optional, {:compound, [
      dimension: :string,
      location: :pos
    ]}},
    portal_cooldown: :varint,
    enforces_secure_chat: :bool,
  ]
  Macros.defpacket_clientbound :synchronize_player_position, 0x40, 767, [
    x: :double,
    y: :double,
    z: :double,
    yaw: :float,
    pitch: :float,
    flags: :byte,
    teleport_id: :varint
  ]

  Macros.defpacket_serverbound :confirm_teleportation, 0x00, 767, [teleport_id: :varint]
  Macros.defpacket_serverbound :set_player_position, 0x1A, 767, [
    x: :double,
    feet_y: :double,
    z: :double,
    on_ground: :bool
  ]
  Macros.defpacket_serverbound :set_player_position_and_rotation, 0x1B, 767, [
    x: :double,
    feet_y: :double,
    z: :double,
    yaw: :float,
    pitch: :float,
    on_ground: :bool
  ]
  Macros.defpacket_serverbound :set_player_rotation, 0x1C, 767, [
    yaw: :float,
    pitch: :float,
    on_ground: :bool # I don't understand their obsession with this...
  ]

  def handle(%{packet_type: :confirm_teleportation, teleport_id: id}, 767, state) do
    Amethyst.API.Game.accept_teleport(state[:game], id)
    :ok
  end

  def handle(%{packet_type: :set_player_position_and_rotation, x: x, feet_y: y, z: z, yaw: yaw, pitch: pitch, on_ground: _ground}, 767, state) do
    # I don't know why we would ever trust on_ground here... the server computes that
    Amethyst.API.Game.player_position(state[:game], {x, y, z})
    Amethyst.API.Game.player_rotation(state[:game], {yaw, pitch})
    :ok
  end
  def handle(%{packet_type: :set_player_position, x: x, feet_y: y, z: z, on_ground: _ground}, 767, state) do
    # I don't know why we would ever trust on_ground here... the server computes that
    Amethyst.API.Game.player_position(state[:game], {x, y, z})
    :ok
  end
  def handle(%{packet_type: :set_player_rotation, yaw: yaw, pitch: pitch, on_ground: _ground}, 767, state) do
    # I don't know why we would ever trust on_ground here... the server computes that
    Amethyst.API.Game.player_rotation(state[:game], {yaw, pitch})
    :ok
  end

  def disconnect(reason) do
    %{
      packet_type: :disconnect,
      reason: {:compound, %{
        "text" => {:string, reason}
      }}
    }
  end
end
